use super::Keyfork;
use clap::{builder::PossibleValue, Parser, Subcommand, ValueEnum};
use std::{
    io::{stdin, stdout, Read, Write},
    path::{Path, PathBuf},
};

const COULD_NOT_DETERMINE_FORMAT: &str = "could not determine format, try including --format";

#[derive(Debug, Clone)]
enum Format {
    OpenPGP(OpenPGP),
    P256(P256),
}

impl ValueEnum for Format {
    fn value_variants<'a>() -> &'a [Self] {
        &[Self::OpenPGP(OpenPGP), Self::P256(P256)]
    }

    fn to_possible_value(&self) -> Option<PossibleValue> {
        Some(match self {
            Format::OpenPGP(_) => PossibleValue::new("openpgp"),
            Format::P256(_) => PossibleValue::new("p256"),
        })
    }
}

trait ShardExec {
    fn split(
        &self,
        threshold: u8,
        max: u8,
        key_discovery: impl AsRef<Path>,
        secret: &[u8],
        output: &mut impl Write,
    ) -> Result<(), Box<dyn std::error::Error>>;

    fn combine<T>(
        &self,
        key_discovery: Option<T>,
        input: impl Read + Send + Sync,
        output: &mut impl Write,
    ) -> Result<(), Box<dyn std::error::Error>>
    where
        T: AsRef<Path>;

    fn decrypt<T>(
        &self,
        key_discovery: Option<T>,
        input: impl Read + Send + Sync,
    ) -> Result<(), Box<dyn std::error::Error>>
    where
        T: AsRef<Path>;
}

#[derive(Clone, Debug)]
struct OpenPGP;

impl ShardExec for OpenPGP {
    fn split(
        &self,
        threshold: u8,
        max: u8,
        key_discovery: impl AsRef<Path>,
        secret: &[u8],
        output: &mut impl Write,
    ) -> Result<(), Box<dyn std::error::Error>> {
        // Get certs and input
        let certs = keyfork_shard::openpgp::discover_certs(key_discovery.as_ref())?;
        assert_eq!(
            certs.len(),
            max.into(),
            "cert count {} != max {max}",
            certs.len()
        );
        keyfork_shard::openpgp::split(threshold, certs, secret, output).map_err(Into::into)
    }

    fn combine<T>(
        &self,
        key_discovery: Option<T>,
        input: impl Read + Send + Sync,
        output: &mut impl Write,
    ) -> Result<(), Box<dyn std::error::Error>>
    where
        T: AsRef<Path>,
    {
        let certs = key_discovery
            .map(|kd| keyfork_shard::openpgp::discover_certs(kd.as_ref()))
            .transpose()?
            .unwrap_or(vec![]);

        let mut encrypted_messages = keyfork_shard::openpgp::parse_messages(input)?;
        let encrypted_metadata = encrypted_messages
            .pop_front()
            .expect("any pgp encrypted message");

        let mut bytes = vec![];

        keyfork_shard::openpgp::combine(
            certs,
            &encrypted_metadata,
            encrypted_messages.into(),
            &mut bytes,
        )?;

        write!(output, "{}", smex::encode(&bytes))?;

        Ok(())
    }

    fn decrypt<T>(
        &self,
        key_discovery: Option<T>,
        input: impl Read + Send + Sync,
    ) -> Result<(), Box<dyn std::error::Error>>
    where
        T: AsRef<Path>,
    {
        let certs = key_discovery
            .map(|kd| keyfork_shard::openpgp::discover_certs(kd.as_ref()))
            .transpose()?
            .unwrap_or(vec![]);

        let mut encrypted_messages = keyfork_shard::openpgp::parse_messages(input)?;
        let encrypted_metadata = encrypted_messages
            .pop_front()
            .expect("any pgp encrypted message");

        keyfork_shard::openpgp::decrypt(
            &certs,
            &encrypted_metadata,
            encrypted_messages.make_contiguous(),
        )?;
        Ok(())
    }
}

#[derive(Clone, Debug)]
struct P256;

#[derive(Subcommand, Clone, Debug)]
pub enum ShardSubcommands {
    /// Split a hex-encoded secret from input into multiple shares, using Shamir's Secret Sharing.
    ///
    /// The shares are encrypted once per key, with keys discovered either on-system or by
    /// prompting for hardware interactions. Metadata about decrypting keys is then stored and
    /// encrypted to all keys, to ensure any key that holds a share can then be used to begin the
    /// process of combining keys.
    Split {
        /// The amount of shares required to recombine a secret.
        #[arg(long)]
        threshold: u8,

        /// The total amount of shares to generate.
        #[arg(long)]
        max: u8,

        /// The path to discover public keys from.
        key_discovery: PathBuf,
    },

    /// Decrypt a single share and re-encrypt it to an ephemeral symmetric key using mnemonic-based
    /// prompts. The mnemonics can be sent over insecure channels.
    Transport {
        /// The path to discover private keys from.
        key_discovery: Option<PathBuf>,
    },

    /// Combine multiple encrypted shares into a hex-encoded secret, printed to stdout.
    ///
    /// This command only accepts input from `keyfork shard split`, and is dependent on the format
    /// used when splitting. Metadata is encrypted to all keys that may hold a share, so when using
    /// hardware metadata discovery, any hardware key used to split may be used to decrypt metadata
    /// used to combine.
    Combine {
        /// The path to discover private keys from.
        key_discovery: Option<PathBuf>,
    },
}

impl ShardSubcommands {
    pub fn handle(
        &self,
        shard: &Shard,
        _keyfork: &Keyfork,
    ) -> Result<(), Box<dyn std::error::Error>> {
        let mut stdin = stdin();
        let mut stdout = stdout();
        let mut input = String::new();
        stdin.read_to_string(&mut input)?;
        let mut format = shard.format.clone();
        // bang sandwich macro fun
        if input.contains("BEGIN PGP MESSAGE") && !matches!(self, ShardSubcommands::Split { .. }) {
            let _ = format.insert(Format::OpenPGP(OpenPGP));
        }
        match self {
            ShardSubcommands::Split {
                threshold,
                max,
                key_discovery,
            } => {
                assert!(threshold <= max, "threshold {threshold} <= max {max}");
                let secret = smex::decode(input.trim())?;
                match format {
                    Some(Format::OpenPGP(o)) => {
                        o.split(*threshold, *max, key_discovery, &secret, &mut stdout)
                    }
                    Some(Format::P256(_p)) => {
                        todo!()
                    }
                    None => panic!("{COULD_NOT_DETERMINE_FORMAT}"),
                }
            }
            ShardSubcommands::Transport { key_discovery } => match format {
                Some(Format::OpenPGP(o)) => o.decrypt(key_discovery.as_ref(), input.as_bytes()),
                Some(Format::P256(_p)) => todo!(),
                None => panic!("{COULD_NOT_DETERMINE_FORMAT}"),
            },
            ShardSubcommands::Combine { key_discovery } => match format {
                Some(Format::OpenPGP(o)) => {
                    o.combine(key_discovery.as_ref(), input.as_bytes(), &mut stdout)
                }
                Some(Format::P256(_p)) => {
                    todo!()
                }
                None => panic!("{COULD_NOT_DETERMINE_FORMAT}"),
            },
        }
    }
}

#[derive(Parser, Debug, Clone)]
pub struct Shard {
    /// Which format to use for encoding/encrypting and decoding/decrypting shares.
    #[arg(long, value_enum, global = true)]
    format: Option<Format>,

    #[command(subcommand)]
    pub command: ShardSubcommands,
}
